﻿using System;
using System.Windows;
using System.Windows.Media.Animation;

namespace TestProject
{
    /// <summary>
    /// 반대 방향 애니메이션
    /// </summary>
    public class BackDirectionAnimation : DoubleAnimationBase
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Enumeration
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 모서리 동작 타입 - EdgeBehaviorType

        /// <summary>
        /// 모서리 동작 타입
        /// </summary>
        public enum EdgeBehaviorType
        {
            /// <summary>
            /// EASE IN
            /// </summary>
            EaseIn,

            /// <summary>
            /// EASE OUT
            /// </summary>
            EaseOut,

            /// <summary>
            /// EASE IN/OUT
            /// </summary>
            EaseInOut
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Dependency Property
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 모서리 동작 속성 - EdgeBehaviorProperty

        /// <summary>
        /// 모서리 동작 속성
        /// </summary>
        public static readonly DependencyProperty EdgeBehaviorProperty = DependencyProperty.Register
        (
            "EdgeBehavior",
            typeof(EdgeBehaviorType),
            typeof(BackDirectionAnimation),
            new PropertyMetadata(EdgeBehaviorType.EaseIn)
        );

        #endregion
        #region 진폭 속성 - AmplitudeProperty

        /// <summary>
        /// 진폭 속성
        /// </summary>
        public static readonly DependencyProperty AmplitudeProperty = DependencyProperty.Register
        (
            "Amplitude",
            typeof(double),
            typeof(BackDirectionAnimation),
            new PropertyMetadata(4.0)
        );

        #endregion
        #region 억제 속성 - SuppressionProperty

        /// <summary>
        /// 억제 속성
        /// </summary>
        public static readonly DependencyProperty SuppressionProperty = DependencyProperty.Register
        (
            "Suppression",
            typeof(double),
            typeof(BackDirectionAnimation),
            new PropertyMetadata(2.0)
        );

        #endregion
        #region 시작 값 속성 - FromProperty

        /// <summary>
        /// 시작 값 속성
        /// </summary>
        public static readonly DependencyProperty FromProperty = DependencyProperty.Register
        (
            "From",
            typeof(double?),
            typeof(BackDirectionAnimation),
            new PropertyMetadata(null)
        );

        #endregion
        #region 종료 값 속성 - ToProperty

        /// <summary>
        /// 종료 값 속성
        /// </summary>
        public static readonly DependencyProperty ToProperty = DependencyProperty.Register
        (
            "To",
            typeof(double?),
            typeof(BackDirectionAnimation),
            new PropertyMetadata(null)
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 모서리 동작 - EdgeBehavior

        /// <summary>
        /// 모서리 동작
        /// </summary>
        public EdgeBehaviorType EdgeBehavior
        {
            get
            {
                return (EdgeBehaviorType)GetValue(EdgeBehaviorProperty);
            }
            set
            {
                SetValue(EdgeBehaviorProperty, value);
            }
        }

        #endregion
        #region 진폭 - Amplitude

        /// <summary>
        /// 진폭
        /// </summary>
        public double Amplitude
        {
            get
            {
                return (double)GetValue(AmplitudeProperty);
            }
            set
            {
                SetValue(AmplitudeProperty, value);
            }
        }

        #endregion
        #region 억제 - Suppression

        /// <summary>
        /// 억제
        /// </summary>
        public double Suppression
        {
            get
            {
                return (double)GetValue(SuppressionProperty);
            }
            set
            {
                SetValue(SuppressionProperty, value);
            }
        }

        #endregion
        #region 시작 값 - From

        /// <summary>
        /// 시작 값
        /// </summary>
        public double? From
        {
            get
            {
                return (double?)GetValue(FromProperty);
            }
            set
            {
                SetValue(FromProperty, value);
            }
        }

        #endregion
        #region 종료 값 - To

        /// <summary>
        /// 종료 값
        /// </summary>
        public double? To
        {
            get
            {
                return (double?)GetValue(ToProperty);
            }
            set
            {
                SetValue(ToProperty, value);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 인스턴스 생성하기 (코어) - CreateInstanceCore()

        /// <summary>
        /// 인스턴스 생성하기 (코어)
        /// </summary>
        /// <returns></returns>
        protected override Freezable CreateInstanceCore() => new BackDirectionAnimation();

        #endregion
        #region 현재 값 구하기 (코어) - GetCurrentValueCore(defaultSourceValue, defaultTargetValue, clock)

        /// <summary>
        /// 현재 값 구하기 (코어)
        /// </summary>
        /// <param name="defaultSourceValue">디폴트 시작 값</param>
        /// <param name="defaultTargetValue">디폴트 종료 값</param>
        /// <param name="clock">애니메이션 클럭</param>
        /// <returns>현재 값</returns>
        protected override double GetCurrentValueCore(double defaultSourceValue, double defaultTargetValue, AnimationClock clock)
        {
            double value;

            var start = From ?? defaultSourceValue;
            var delta = To - start ?? defaultSourceValue - start;

            switch(EdgeBehavior)
            {
                case EdgeBehaviorType.EaseIn :

                    value = EaseIn(clock.CurrentProgress.Value, start, delta, Amplitude, Suppression);

                    break;

                case EdgeBehaviorType.EaseOut :

                    value = EaseOut(clock.CurrentProgress.Value, start, delta, Amplitude, Suppression);

                    break;

                default :

                    value = EaseInOut(clock.CurrentProgress.Value, start, delta, Amplitude, Suppression);

                    break;
            }

            return value;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region EASE IN 처리하기 - EaseIn(timeFraction, start, delta, amplitude, suppression)

        /// <summary>
        /// EASE IN 처리하기
        /// </summary>
        /// <param name="timeFraction">시간 분할</param>
        /// <param name="start">시작 값</param>
        /// <param name="delta">증분</param>
        /// <param name="amplitude">진폭</param>
        /// <param name="suppression">억제</param>
        /// <returns>처리 결과</returns>
        private double EaseIn(double timeFraction, double start, double delta, double amplitude, double suppression)
        {
            double frequency = 0.5;

            double value = Math.Pow(1 - timeFraction, suppression) * amplitude * Math.Sin(2 * Math.PI * timeFraction*frequency) * -1 + timeFraction;

            value = value * delta;

            value += start;

            return value;
        }

        #endregion
        #region EASE OUT 처리하기 - EaseOut(timeFraction, start, delta, amplitude, suppression)

        /// <summary>
        /// EASE OUT 처리하기
        /// </summary>
        /// <param name="timeFraction">시간 분할</param>
        /// <param name="start">시작 값</param>
        /// <param name="delta">증분</param>
        /// <param name="amplitude">진폭</param>
        /// <param name="suppression">억제</param>
        /// <returns>처리 결과</returns>
        private double EaseOut(double timeFraction, double start, double delta, double amplitude, double suppression)
        {
            double frequency = 0.5;

            double value = Math.Pow(timeFraction, suppression) * amplitude * Math.Sin(2 * Math.PI * timeFraction * frequency) + timeFraction;

            value = value * delta;

            value += start;

            return value;
        }

        #endregion
        #region EASE IN/OUT 처리하기 - EaseInOut(timeFraction, start, delta, amplitude, suppression)

        /// <summary>
        /// EASE IN/OUT 처리하기
        /// </summary>
        /// <param name="timeFraction">시간 분할</param>
        /// <param name="start">시작 값</param>
        /// <param name="delta">증분</param>
        /// <param name="amplitude">진폭</param>
        /// <param name="suppression">억제</param>
        /// <returns>처리 결과</returns>
        private double EaseInOut(double timeFraction, double start, double delta, double amplitude, double suppression)
        {
            double returnValue;

            if(timeFraction <= 0.5)
            {
                return EaseIn(timeFraction * 2, start, delta / 2, amplitude, suppression);
            }

            returnValue = EaseOut((timeFraction - 0.5) * 2, start, delta / 2, amplitude, suppression);

            returnValue += delta / 2;

            return returnValue;
        }

        #endregion
    }
}