﻿using System;
using System.Windows;
using System.Windows.Media.Animation;

namespace TestProject
{
    /// <summary>
    /// 바운스 애니메이션
    /// </summary>
    public class BounceAnimation : DoubleAnimationBase
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Enumeration
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 모서리 동작 타입 - EdgeBehaviorType

        /// <summary>
        /// 모서리 동작 타입
        /// </summary>
        public enum EdgeBehaviorType
        {
            /// <summary>
            /// EASE IN
            /// </summary>
            EaseIn,

            /// <summary>
            /// EASE OUT
            /// </summary>
            EaseOut,

            /// <summary>
            /// EASE IN/OUT
            /// </summary>
            EaseInOut
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Dependency Property
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 모서리 동작 속성 - EdgeBehaviorProperty

        /// <summary>
        /// 모서리 동작 속성
        /// </summary>
        public static readonly DependencyProperty EdgeBehaviorProperty = DependencyProperty.Register
        (
            "EdgeBehavior",
            typeof(EdgeBehaviorType),
            typeof(BounceAnimation),
            new PropertyMetadata(EdgeBehaviorType.EaseInOut)
        );

        #endregion
        #region 바운스 카운트 속성 - BounceCountProperty

        /// <summary>
        /// 바운스 카운트 속성
        /// </summary>
        public static readonly DependencyProperty BounceCountProperty = DependencyProperty.Register
        (
            "BounceCount",
            typeof(int),
            typeof(BounceAnimation),
            new PropertyMetadata(5)
        );

        #endregion
        #region 탄력도 속성 - BouncinessProperty

        /// <summary>
        /// 탄력도 속성
        /// </summary>
        public static readonly DependencyProperty BouncinessProperty = DependencyProperty.Register
        (
            "Bounciness",
            typeof(double),
            typeof(BounceAnimation),
            new PropertyMetadata(3.0)
        );

        #endregion
        #region 시작 값 속성 - FromProperty

        /// <summary>
        /// From 시작 값
        /// </summary>
        public static readonly DependencyProperty FromProperty = DependencyProperty.Register
        (
            "From",
            typeof(double?),
            typeof(BounceAnimation),
            new PropertyMetadata(null)
        );

        #endregion
        #region 종료 값 속성 - ToProperty

        /// <summary>
        /// 종료 값 속성
        /// </summary>
        public static readonly DependencyProperty ToProperty = DependencyProperty.Register
        (
            "To",
            typeof(double?),
            typeof(BounceAnimation),
            new PropertyMetadata(null)
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 모서리 동작 - EdgeBehavior

        /// <summary>
        /// 모서리 동작
        /// </summary>
        public EdgeBehaviorType EdgeBehavior
        {
            get
            {
                return (EdgeBehaviorType)GetValue(EdgeBehaviorProperty);
            }
            set
            {
                SetValue(EdgeBehaviorProperty, value);
            }
        }

        #endregion
        #region 바운스 카운트 - BounceCount

        /// <summary>
        /// 바운스 카운트
        /// </summary>
        public int BounceCount
        {
            get
            {
                return (int)GetValue(BounceCountProperty);
            }
            set
            {
                if(value > 0)
                {
                    SetValue(BounceCountProperty, value);
                }
                else
                {
                    throw new ArgumentException($"바운스 카운트를 설정할 수 없습니다 : {value}");
                }
            }
        }

        #endregion
        #region 탄력도 - Bounciness

        /// <summary>
        /// 탄력도
        /// </summary>
        public double Bounciness
        {
            get
            {
                return (double)GetValue(BouncinessProperty);
            }
            set
            {
                if(value > 0)
                {
                    SetValue(BouncinessProperty, value);
                }
                else
                {
                    throw new ArgumentException($"탄력도를 설정할 수 없습니다 : {value}");
                }
            }
        }

        #endregion
        #region 시작값 - From

        /// <summary>
        /// 시작값
        /// </summary>
        public double? From
        {
            get
            {
                return (double?)GetValue(FromProperty);
            }
            set
            {
                SetValue(FromProperty, value);
            }
        }

        #endregion
        #region 종료값 - To

        /// <summary>
        /// 종료값
        /// </summary>
        public double? To
        {
            get
            {
                return (double?)GetValue(ToProperty);
            }
            set
            {
                SetValue(ToProperty, value);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 인스턴스 생성하기 (코어) - CreateInstanceCore()

        /// <summary>
        /// 인스턴스 생성하기 (코어)
        /// </summary>
        /// <returns>생성 인스턴스</returns>
        protected override Freezable CreateInstanceCore() => new BounceAnimation();

        #endregion
        #region 현재 값 구하기 (코어) - GetCurrentValueCore(defaultSourceValue, defaultTargetValue, clock)

        /// <summary>
        /// 현재 값 구하기 (코어)
        /// </summary>
        /// <param name="defaultSourceValue">디폴트 소스 값</param>
        /// <param name="defaultTargetValue">디폴트 타겟 값</param>
        /// <param name="clock">애니메이션 클럭</param>
        /// <returns>현재 값</returns>
        protected override double GetCurrentValueCore(double defaultSourceValue, double defaultTargetValue, AnimationClock clock)
        {
            double value;

            double start = From ?? defaultSourceValue;
            double delta = To - start ?? defaultSourceValue - start;

            switch(EdgeBehavior)
            {
                case EdgeBehaviorType.EaseIn :

                    value = EaseIn(clock.CurrentProgress.Value, start, delta, Bounciness, BounceCount);

                    break;

                case EdgeBehaviorType.EaseOut :

                    value = EaseOut(clock.CurrentProgress.Value, start, delta, Bounciness, BounceCount);

                    break;

                default :

                    value = EaseInOut(clock.CurrentProgress.Value, start, delta, Bounciness, BounceCount);

                    break;
            }

            return value;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region EASE OUT 처리하기 - EaseOut(timeFraction, start, delta, bounciness, bounceCount)

        /// <summary>
        /// EASE OUT 처리하기
        /// </summary>
        /// <param name="timeFraction">시간 분할</param>
        /// <param name="start">시작 값</param>
        /// <param name="delta">증분</param>
        /// <param name="bounciness">탄력도</param>
        /// <param name="bounceCount">바운스 카운트</param>
        /// <returns>처리 결과</returns>
        private double EaseOut(double timeFraction, double start, double delta, double bounciness, int bounceCount)
        {
            double value = Math.Abs(Math.Pow(1 - timeFraction, bounciness) * Math.Cos(2 * Math.PI * timeFraction * bounceCount));

            value = delta - (value * delta);

            value += start;

            return value;
        }

        #endregion
        #region EASE IN 처리하기 - EaseIn(timeFraction, start, delta, bounciness, bounceCount)

        /// <summary>
        /// EASE IN 처리하기
        /// </summary>
        /// <param name="timeFraction">시간 분할</param>
        /// <param name="start">시작 값</param>
        /// <param name="delta">증분</param>
        /// <param name="bounciness">탄력도</param>
        /// <param name="bounceCount">바운스 카운트</param>
        /// <returns>처리 결과</returns>
        private double EaseIn(double timeFraction, double start, double delta, double bounciness, int bounceCount)
        {
            double value = Math.Abs(Math.Pow(timeFraction, bounciness) * Math.Cos(2 * Math.PI * timeFraction * bounceCount));

            value = value * delta;

            value += start;

            return value;
        }

        #endregion
        #region EASE IN/OUT 처리하기 - EaseInOut(timeFraction, start, delta, bounciness, bounceCount)

        /// <summary>
        /// EASE IN/OUT 처리하기
        /// </summary>
        /// <param name="timeFraction">시간 분할</param>
        /// <param name="start">시작 값</param>
        /// <param name="delta">증분</param>
        /// <param name="bounciness">탄력도</param>
        /// <param name="bounceCount">바운스 카운트</param>
        /// <returns>처리 결과</returns>
        private double EaseInOut(double timeFraction, double start, double delta, double bounciness, int bounceCount)
        {
            double value;

            if(timeFraction <= 0.5)
            {
                value = EaseIn(timeFraction * 2, start, delta / 2, bounciness, bounceCount);
            }
            else
            {
                value = EaseOut((timeFraction - 0.5) * 2, start, delta / 2, bounciness, bounceCount);

                value += delta / 2;
            }

            return value;
        }

        #endregion
    }
}