﻿using System;
using System.Windows;
using System.Windows.Media.Animation;

namespace TestProject
{
    /// <summary>
    /// 지수 애니메이션
    /// </summary>
    public class ExponentialAnimation : DoubleAnimationBase
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Enumeration
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 모서리 동작 타입 - EdgeBehaviorType

        /// <summary>
        /// 모서리 동작 타입
        /// </summary>
        public enum EdgeBehaviorType
        {
            /// <summary>
            /// EASE IN
            /// </summary>
            EaseIn,

            /// <summary>
            /// EASE OUT
            /// </summary>
            EaseOut,

            /// <summary>
            /// EASE IN/OUT
            /// </summary>
            EaseInOut
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Dependency Property
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 모서리 동작 속성 - EdgeBehaviorProperty

        /// <summary>
        /// 모서리 동작 속성
        /// </summary>
        public static readonly DependencyProperty EdgeBehaviorProperty = DependencyProperty.Register
        (
            "EdgeBehavior",
            typeof(EdgeBehaviorType),
            typeof(ExponentialAnimation),
            new PropertyMetadata(EdgeBehaviorType.EaseIn)
        );

        #endregion
        #region 거듭 제곱 속성 - PowerProperty

        /// <summary>
        /// 거듭 제곱 속성
        /// </summary>
        public static readonly DependencyProperty PowerProperty = DependencyProperty.Register
        (
            "Power",
            typeof(double),
            typeof(ExponentialAnimation),
            new PropertyMetadata(2.0)
        );

        #endregion
        #region 시작 값 속성 - FromProperty

        /// <summary>
        /// 시작 값 속성
        /// </summary>
        public static readonly DependencyProperty FromProperty = DependencyProperty.Register
        (
            "From",
            typeof(double?),
            typeof(ExponentialAnimation),
            new PropertyMetadata(null)
        );

        #endregion
        #region 종료 값 속성 - ToProperty

        /// <summary>
        /// 종료 값 속성
        /// </summary>
        public static readonly DependencyProperty ToProperty = DependencyProperty.Register
        (
            "To",
            typeof(double?),
            typeof(ExponentialAnimation),
            new PropertyMetadata(null)
        );

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 모서리 동작 - EdgeBehavior

        /// <summary>
        /// 모서리 동작
        /// </summary>
        public EdgeBehaviorType EdgeBehavior
        {
            get
            {
                return (EdgeBehaviorType)GetValue(EdgeBehaviorProperty);
            }
            set
            {
                SetValue(EdgeBehaviorProperty, value);
            }
        }

        #endregion
        #region 거듭 제곱 - Power

        /// <summary>
        /// 거듭 제곱
        /// </summary>
        public double Power
        {
            get
            {
                return (double)GetValue(PowerProperty);
            }
            set
            {
                if(value > 0.0)
                {
                    SetValue(PowerProperty, value);
                }
                else
                {
                    throw new ArgumentException($"0보다 작은 거듭제곱 값을 설정할 수 없습니다 : {value}");
                }
            }
        }

        #endregion
        #region 시작 값 - From

        /// <summary>
        /// 시작 값
        /// </summary>
        public double? From
        {
            get
            {
                return (double?)GetValue(FromProperty);
            }
            set
            {
                SetValue(FromProperty, value);
            }
        }

        #endregion
        #region 종료 값 - To

        /// <summary>
        /// 종료 값
        /// </summary>
        public double? To
        {
            get
            {
                return (double?)GetValue(ToProperty);
            }
            set
            {
                SetValue(ToProperty, value);
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 인스턴스 생성하기 (코어) - CreateInstanceCore()

        /// <summary>
        /// 인스턴스 생성하기 (코어)
        /// </summary>
        /// <returns>생성 인스턴스</returns>
        protected override Freezable CreateInstanceCore() => new ExponentialAnimation();

        #endregion
        #region 현재 값 구하기 (코어) - GetCurrentValueCore(defaultSourceValue, defaultTargetValue, clock)

        /// <summary>
        /// 현재 값 구하기 (코어)
        /// </summary>
        /// <param name="defaultSourceValue">디폴트 소스 값</param>
        /// <param name="defaultTargetValue">디폴트 타겟 값</param>
        /// <param name="clock">애니메이션 클럭</param>
        /// <returns>현재 값</returns>
        protected override double GetCurrentValueCore(double defaultSourceValue, double defaultTargetValue, AnimationClock clock)
        {
            double value;

            double start = From ?? defaultSourceValue;
            double delta = To - start ?? defaultSourceValue - start;

            switch(EdgeBehavior)
            {
                case EdgeBehaviorType.EaseIn :

                    value = EaseIn(clock.CurrentProgress.Value, start, delta, Power);

                    break;

                case EdgeBehaviorType.EaseOut :

                    value = EaseOut(clock.CurrentProgress.Value, start, delta, Power);

                    break;

                default :

                    value = EaseInOut(clock.CurrentProgress.Value, start, delta, Power);

                    break;
            }

            return value;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region EASE IN 처리하기 - EaseIn(timeFraction, start, delta, power)

        /// <summary>
        /// EASE IN 처리하기
        /// </summary>
        /// <param name="timeFraction">시간 분할</param>
        /// <param name="start">시작 값</param>
        /// <param name="delta">증분</param>
        /// <param name="power">거듭 제곱</param>
        /// <returns>처리 결과</returns>
        private double EaseIn(double timeFraction, double start, double delta, double power)
        {
            double value = Math.Pow(timeFraction, power);

            value *= delta;

            value = value + start;

            return value;
        }

        #endregion
        #region EASE OUT 처리하기 - EaseOut(timeFraction, start, delta, power)

        /// <summary>
        /// EASE OUT 처리하기
        /// </summary>
        /// <param name="timeFraction">시간 분할</param>
        /// <param name="start">시작 값</param>
        /// <param name="delta">증분</param>
        /// <param name="power">거듭 제곱</param>
        /// <returns>처리 결과</returns>
        private double EaseOut(double timeFraction, double start, double delta, double power)
        {
            double value = Math.Pow(timeFraction, 1 / power);

            value *= delta;

            value = value + start;

            return value;
        }

        #endregion
        #region EASE IN/OUT 처리하기 - EaseInOut(timeFraction, start, delta, power)

        /// <summary>
        /// EASE IN/OUT 처리하기
        /// </summary>
        /// <param name="timeFraction">시간 분할</param>
        /// <param name="start">시작 값</param>
        /// <param name="delta">증분</param>
        /// <param name="power">거듭 제곱</param>
        /// <returns>처리 결과</returns>
        private double EaseInOut(double timeFraction, double start, double delta, double power)
        {
            double value;

            if(timeFraction <= 0.5)
            {
                value = EaseOut(timeFraction * 2, start, delta / 2, power);
            }
            else
            {
                value = EaseIn((timeFraction - 0.5) * 2, start, delta / 2, power);

                value += delta / 2;
            }

            return value;
        }

        #endregion
    }
}